;************************************************************************************************
; Fast AVR FFT - main FFT routines
;************************************************************************************************
; Fast FFT routines in 16-bit fixed point resolution. Cooley-Tukey DIT radix 2 algorithm.  
; All functions uses common global complex work array:
;   int16_t FFTwarr[FFTN*2]; 
; where FFTN is FFT size, memory organisation in array:
;   re{0},im{0},re{1},im{1},...,re{FFTN-1},im{FFTN-1}
; Routines should work for FFTN={32,64,128,256,512}, lower resolutions are also possible, 
; but there are no lookups generated for them.
;
; Usage:
; ------
; 1. fill im{} items of FFTwarr[] with signal to process
; 2. call function FFT_input()
;     - this function does bit reversion and applies selected window      
;     - then optionally duplicates re{} to im{} to increase amplitude by 2^0.5 (better SNR) 
; 3. call FFT_dit_r2() to do main FFT DIT radix 2 transformation
;     - complex results are stored in FFTwarr and are already in correct order
; 4. call FFT_out_lin() or FFT_out_log() to create linear or log scale amplitude spectrum from 
;    first half of complex spectrum, amplitudes "r" are stored into FFTwarr[] in this order:
;      FFTwarr[0]=r[0],FFTwarr[1]=r[1],FFTwarr[2]=r[2],... 
;
; Options (FFT.h):
; ----------------
; FFTN     - FFT size	     
; FFTW     - window to apply
; FFTdupim - duplicate input signal to both re{} and im{} part of FFTwarr[]
;          - this increases input signal amplitude by 2^0.5 so it'll have better SNR
; FFTprec  - increases algorithm precision (lower algorithm noise), needs some more CPU power
; FFTopt   - increases algorithm performance (bigger code size)
; FFTfsqrt - uses fast lookup sqrt() insted of precise one
;
; Files:
; ------
; Whole FFT library consist of several files:
; 1. FFT_routines.s - main FFT algorithm routines
; 2. FFT_macros.inc - some higher level FFT related macros
; 3. FFT_tables.inc - FFT lookups
; 4. FFT_math.s     - FFT related arithmetics
; 5. FFT.c          - global FFTwarr[] array and some demo functions
; 6. FFT.h          - common FFT library header file, algorithm parameters
; 7. macros.inc     - basic assembler macros, neccessary for all other assembler modules
;
;
;
; Measured execution times for FFTfsqrt=0, FFTprec=0, FFTdupim=1, XTAL=16MHz:
;
; FFTopt=0, linear output:
; FFTN   input  transformation  output  total   performance
; [-]    [ms]   [ms]            [ms]    [ms]    [kp/s]
; 32     0.118   0.707          0.601    1.426  22.440
; 64     0.235   1.670          1.203    3.108  20.592
; 128    0.485   3.855          2.410    6.750  18.963
; 256    0.969   8.745          4.821   14.535  17.613
; 512    1.937  19.567          9.643   31.147  16.438
;
; FFTopt=1, linear output:
; FFTN   input  transformation  output  total   performance
; [-]    [ms]   [ms]            [ms]    [ms]    [kp/s]
; 32     0.118   0.500          0.601    1.218  26.273
; 64     0.235   1.251          1.204    2.690  23.796
; 128    0.485   3.012          2.409    5.906  21.673
; 256    0.969   7.057          4.820   12.845  19.930
; 512    1.937  16.190          9.643   27.769  18.437
;
; FFTopt=0, log scale output:
; FFTN   input  transformation  output   total   performance
; [-]    [ms]   [ms]            [ms]     [ms]    [kp/s]
; 32     0.118   0.707           0.655    1.480  21.622
; 64     0.235   1.670           1.305    3.210  19.935
; 128    0.485   3.855           2.613    6.953  18.408
; 256    0.969   8.745           5.227   14.941  17.143
; 512    1.937  19.567          10.458   31.962  16.019
;
; FFTopt=1, log scale output:
; FFTN   input  transformation  output   total   performance
; [-]    [ms]   [ms]            [ms]     [ms]    [kp/s]
; 32     0.118   0.500           0.655    1.272  25.157
; 64     0.235   1.250           1.306    2.791  22.931
; 128    0.485   3.012           2.609    6.106  20.963
; 256    0.969   7.056           5.216   13.241  19.334
; 512    1.937  16.190          10.431   28.558  17.928
;
; Performance for FFTprec=1 is 12% lower or so. 
;
; (c) 2012, Stanislav Maslan, s.maslan@seznam.cz
; http://www.elektronika.kvalitne.cz/ATMEL/necoteorie/transformation/AVRFFT/AVRFFT.html
;************************************************************************************************
#include <avr/io.h>

#include "FFT.h"

// general macros
#include "macros.inc"

// FFT related macros
#include "FFT_macros.inc"

// FFT lookup tables
#include "FFT_tables.inc"


;================================================================================================
; Apply window and bitreversion to input data source data: 
;   re{FFTwarr[]} [fx 1.15]
;
; Algorithm description: 
;   step 1: re{FFTwarr[bit_rev(i)]}=im{FFTwarr[i]}*window[i] for i=0...(FFTN-1) [fx 1.15]
;   step 2: im{FFTwarr[i]}=re{FFTwarr[i]} for i=0...(FFTN-1) (re{} to im{} duplication if enabled)
;================================================================================================
	.global	FFT_input
	.func	FFT_input

FFT_input:	push	R2
	pushd	R4
	pushw	R16
	pushw	R28

	clr	R2	;0
	.if	FFTW!=wRECT
	;=== for windows only ===
	ldiw	R30,TFFTwdw	;Z=FFT window
	ldiw	R20,TFFTbrv	;R20:R21=FFT bit reversion table
	ldiw	R28,FFTwarr+2	;Y=FFT work array
	ldiw	R24,FFTN	;FFTN
FFTinpL1:	ldw	R16,Y+	;source=im{*Y++}
	adiw	R28,2
	lpmw	R18,Z+	;window_item=*Z++
		
	.ifdef	FFTWS
	fmuls	R17,R19
	.else
	fmulsu	R17,R19
	.endif
	movw	R6,R0
	fmul	R16,R18
	movw	R4,R0
	adc	R6,R2
	fmulsu	R17,R18
	sbc	R7,R2
	addw	R5,R0
	adc	R7,R2
	.ifdef	FFTWS
	fmulsu	R19,R16	;signed window
	sbc	R7,R2
	.else
	fmul	R19,R16	;unsigned window
	.endif
	addw	R5,R0
	adc	R7,R2	;res <= (un)signed(window_item)*signed(source) [fx 1.15]
		
	movw	R22,R30	;backup Z
	movw	R30,R20	;Z=R20:R21
	.if	FFTN>64
	lpmw	R26,Z+	;br=*Z++
	.else
	lpm	R26,Z+
	clr	R27
	.endif
	movw	R20,R30	;R20:R21=Z
	movw	R30,R22	;restore Z
	subiw	R26,-(FFTwarr)	;X=&FFTwarr[bit_rev(br)]
	stw	X+,R6	;re{*X}=res
	;new item	
	dbnzwh	R24,FFTinpL1
	
	.else
	;=== if no window selected (rectangular) ===
	ldiw	R30,TFFTbrv	;Z=FFT bit reversion table
	ldiw	R28,FFTwarr+2	;Y=FFT work array
	ldiw	R24,FFTN	;FFTN
FFTinpL1:	ldw	R16,Y+	;source=im{*Y++}
	adiw	R28,2		
	.if	FFTN>64
	lpmw	R26,Z+	;br=*Z++
	.else
	lpm	R26,Z+
	clr	R27
	.endif
	subiw	R26,-(FFTwarr)	;X=&FFTwarr[bit_rev(br)]
	stw	X+,R16	;re{*X}=source
	;new item	
	dbnzwh	R24,FFTinpL1
	.endif
	
	.if	FFTdupim
	;duplicate re{} into im{}
	ldiw	R28,FFTwarr
	ldi	R24,low(FFTN/2)
FFTinpL2:	ldw	R16,Y+
	stw	Y+,R16
	ldw	R16,Y+
	stw	Y+,R16
	dbnz	R24,FFTinpL2
	.else
	;clear im{}
	ldi	R16,0
	ldiw	R28,FFTwarr+2
	ldi	R24,low(FFTN/2)
FFTinpL2:	st	Y+,R16
	st	Y+,R16
	adiw	R28,2
	st	Y+,R16
	st	Y+,R16
	adiw	R28,2
	dbnz	R24,FFTinpL2
	.endif
	
	popw	R28
	popw	R16
	popd	R4
	pop	R2
	clr	R1
	ret
	.endfunc



	
;================================================================================================
; FFT DIT radix 2 algorithm for FFTN from 2 upto 512 points FFT, used variables:
;   FFTwarr - FFT work array of FFTN complex values [fx 0.16] 
;   n       - butterfly inputs A and B distance in FFTwarr for actual FFT stage 
;   angf    - angluar frequency for actual FFT stage
;   ang     - actual angle of twiddle factor (Z pointer to sincos lookup)
;
; Uses/changes almost all MCU registers
;
; (c) Stanislav Maslan, 2012, s.maslan@seznam.cz 
;================================================================================================
	.global	FFT_dit_r2
	.func	FFT_dit_r2

FFT_dit_r2:	pushd	R2
	pushd	R6
	pushd	R10
	pushd	R14
	pushw	R28

	ser	R16
	mov	R3,R16	;255
	clr	R2	;0
	ldi	R16,low(FFTN/2)
	mov	R15,R16	;angf=FFTN/2
	clr	R14
	inc	R14	;n=1		
	;====== for each FFT stage ======
FFTditr2E:	;calculate pointers *Y=&A and *X=&B, where A, B are initial butterfly input values  
	ldi	R22,4	;4
	
	ldiw	R28,FFTwarr	;Y=&FFTwarr[0]
	movw	R26,R28
	.if	FFTN>=512
	tst	R14
	brne	PC+(1+1+1)*2
	subi	R27,-2*2
	rjmp	PC+(1+1+2)*2
	.endif
	mul	R14,R22	;n*4
	addw	R26,R0	;X=&FFTwarr[n]
	
	;calculate butterfly step constant 2*n
	.if	FFTN>=512
	tst	R14
	brne	PC+(1+2+1)*2
	ldiw	R24,2*2*2*256
	rjmp	PC+(1+2+1)*2
	.endif
	lslw	R0
	movw	R24,R0	;store it
		
	;sin[], cos[] lookup
	ldiw	R30,TFFTsico	;Z=TFFTsico, Z also represents angle "ang"
		
	
	.if	FFTopt
	;--- simplified butterfly for 0 angle (first butterfly offset) ---
	FFTditr2a0M	;do simplified butterfly
	bst	R14,0
	rjts	FFTditr2E0	;skip rest if n==1
	subiw	R28,(FFTN-1)*2*2	;Y-=(FFTN-1)
	subiw	R26,(FFTN-1)*2*2	;X-=(FFTN-1)
	mul	R15,R22
	addw	R30,R0	;ang+=angf	
	.endif

	
	mov	R12,R14	;i (angle counter)
	.if	FFTopt
	dec	R12
	.endif
	;====== for each twiddle factor angle ======
FFTditr2TA:	lpmw	R20,Z+	;dsin=sin(ang)
	lpmw	R22,Z+	;dcos=cos(ang)
	
	.if	FFTopt
	;--- simplified butterfly for pi/2 angle ---
	cp	R22,R2
	cpc	R23,R2
	brne	FFTditr2S0N
	FFTditr2api2M	;do simplified butterfly if cos()==0
	rjmp	FFTditr2S0
FFTditr2S0N:	
	.endif

	
	mov	R13,R15	;butterfly counter
	;====== for each butterfly ======
	;butterfly algorithm: A_new <= A_old + Fki*B_old
	;                     B_new <= A_old - Fki*B_old
	;where Fki is twiddle factor exp(-j*w*ang) and A, B are inputs/outputs of butterfly         
FFTditr2B:	ldw	R16,X+	;Bre=re{*X}
	ldw	R18,X+	;Bim=im{*X}
	;rotation R=B*Fki in rectagonal form
	maddw	R16,R22,R18,R20,R4,R8,R10,R2	;Rre=Bim*dsin+Bre*dcos [fx 0.16]
	.if	FFTprec
	brpl	PC+(1+4)*2
	add	R4,R3
	adc	R5,R3
	adc	R8,R2
	adc	R9,R2
	.endif
	msubw	R18,R22,R16,R20,R4,R6,R10,R2	;Rim=Bim*dcos-Bre*dsin [fx 0.16]
	.if	FFTprec
	brpl	PC+(1+4)*2
	add	R4,R3
	adc	R5,R3
	adc	R6,R2
	adc	R7,R2
	.endif		
	;finish butterfly - write back new values 
	lddw	R16,Y+0
	asrw	R16	;Are=re{*Y}>>1
	.if	FFTprec
	tst	R17
	brpl	PC+(1+1+1)*2
	adc	R16,R2
	adc	R17,R2
	.endif
	lddw	R18,Y+2
	asrw	R18	;Aim=im{*Y}>>1
	.if	FFTprec
	tst	R19
	brpl	PC+(1+1+1)*2
	adc	R18,R2
	adc	R19,R2
	.endif
	movw	R4,R16
	movw	R10,R18	;backup Are,Aim
	;for A output
	addw	R16,R8
	addw	R18,R6
	stdw	Y+0,R16	;re{*Y}=Are+Rre
	stdw	Y+2,R18	;im{*y}=Aim+Rim
	;for B output
	subw	R4,R8
	subw	R10,R6
	st	-X,R11
	st	-X,R10	;im{*X}=Aim-Rim
	st	-X,R5
	st	-X,R4	;re{*X}=Are-Rre
	;move input pointers to next butterfly inputs
	addw	R28,R24	;Y+=2*n
	addw	R26,R24	;X+=2*n
	djnz	R13,FFTditr2B	;next butterfly
	
	;move input pointers to next butterfly offset 
FFTditr2S0:	subiw	R28,(FFTN-1)*2*2	;Y-=(FFTN-1)
	subiw	R26,(FFTN-1)*2*2	;X-=(FFTN-1)
	
	;new twiddle factor angle
	ldi	R16,4
	mul	R16,R15
	sbiw	R30,2*2
	addw	R30,R0	;ang+=angf
	djnz	R12,FFTditr2TA	;goto next twiddle factor angle
	
FFTditr2E0:	lsl	R14	;n<<=1
	.if	FFTN>=512
	clc
	tst	R15
	brne	PC+(1+1)*2
	sec
	ror	R15
	.else
	lsr	R15	;angf>>=1
	.endif
	rjcc	FFTditr2E	;next FFT stage (log2(FFTN) stages in total)
	;and done!
	
	popw	R28
	popd	R14
	popd	R10
	popd	R6
	popd	R2
	clr	R1
	ret
	.endfunc
	
	
;================================================================================================
; Calculates linear scale amplitude spectrum from first half of FFTwarr[] spectrum 
; Algorithm: 
;   FFTwarr[i]=sqrt(2*(re{FFTwarr[i]}^2 + im{FFTwarr[i]}^2)) [fx 0.16]
;================================================================================================
	.global	FFT_out_lin
	.func	FFT_out_lin
	
FFT_out_lin:	pushd	R2
	pushd	R6
	pusht	R15
	pushw	R28

	clr	R15	;0
	ldiw	R28,FFTwarr	;FFTwarr source 
	ldiw	R26,FFTwarr	;FFTwarr dest
	ldi	R24,low(FFTN/2)	;FFTN/2 (only half of spectrum)
FFToutLinL1:	ldw	R16,Y+	;re=re{*Y}
	ldw	R18,Y+	;im=im{*Y++}
	pwrsw	R16,R2,R15	;re*re
	pwrsw	R18,R6,R15	;im*im
	addd	R2,R6	;re*re+im*im
	;r=sqrt32(re*re+im*im)
	.if	FFTfsqrt
	rcall	FFT_sqrt32f_asm
	.else
	rcall	FFT_sqrt32c_asm
	.endif
	stw	X+,R16	;*X++=r
	djnz	R24,FFToutLinL1
	
	popw	R28
	popt	R15
	popd	R6
	popd	R2
	clr	R1
	ret
	.endfunc
	
	
	
	
	
;================================================================================================
; Calculates log scale amplitude spectrum from transformed FFTwarr data
; Algorithm:
;   r=sqrt((re{FFTwarr[i]}^2 + im{FFTwarr[i]}^2)*2)
;   FFTwarr[i]=20*log10(r/max) [fx 0.16], where max is valu for 0dB output (depends on window) 
;================================================================================================
	.global	FFT_out_log
	.func	FFT_out_log

FFT_out_log:	pushd	R2
	pushd	R6
	pusht	R15
	pushw	R28
	;
	clr	R15	;0
	ldiw	R28,FFTwarr	;FFTwarr source 
	ldiw	R26,FFTwarr	;FFTwarr dest
	ldi	R24,low(FFTN/2)	;FFTN/2 (only half of spectrum)
FFToutLogL1:	ldw	R16,Y+	;re=re{*Y}
	ldw	R18,Y+	;im=im{*Y++}
	pwrsw	R16,R2,R15	;re*re
	pwrsw	R18,R6,R15	;im*im
	addd	R2,R6	;re*re+im*im
	;r=sqrt32(re*re+im*im)
	.if	FFTfsqrt
	rcall	FFT_sqrt32f_asm
	.else
	rcall	FFT_sqrt32c_asm
	.endif
	rcall	FFT_log10f_asm	;r=log10(r/18383) [dB] [fx 8.8]
	stw	X+,R16	;*X++=r
	djnz	R24,FFToutLogL1
	;
	popw	R28
	popt	R15
	popd	R6
	popd	R2
	clr	R1
	ret
	.endfunc
